# Test file only (not up-to-date with gui/app.py) to quickly check if nuitka can build an executable successfully.
# This file is *much* quicker to test because it does not link against ezsam.cli (and thus skips a lot of the heavier ML libraries).

import sys
import threading
import tkinter as tk

import PIL as pil
import customtkinter as ctk
import tkinterdnd2 as dnd

from ezsam.lib.logger import debug, log
from ezsam.lib.tkutils import create_required_stringvars
from ezsam.lib.path import resource_path
from ezsam.lib.preview import get_preview_image
from ezsam.gui.config import (
  DEFAULT_DEBUG,
  DEFAULT_MODEL,
  WIDTH,
  HEIGHT,
  PREVIEW_WIDTH,
  PREVIEW_HEIGHT,
  APPEARANCE_MODE,
  COLOR_THEME,
)
from ezsam.gui.models import HQ_MODEL_PREFIX, MODEL_NAME_TO_TYPE


class App(ctk.CTk, dnd.TkinterDnD.DnDWrapper):
  NAME = 'ezsam'
  WIDTH = WIDTH
  HEIGHT = HEIGHT
  PREVIEW_WIDTH = PREVIEW_WIDTH
  PREVIEW_HEIGHT = PREVIEW_HEIGHT

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.TkdndVersion = dnd.TkinterDnD._require(self)
    self.setup()
    self.layout()
    self.create_widgets()

  def setup(self):
    self.title(App.NAME)
    self.geometry(f'{App.WIDTH}x{App.HEIGHT}')
    self.minsize(App.WIDTH, App.HEIGHT)
    self.protocol('WM_DELETE_WINDOW', self.quit)
    if sys.platform == 'darwin':
      self.bind('<Command-q>', self.quit)
      self.bind('<Command-w>', self.quit)
      self.createcommand('tk::mac::Quit', self.quit)
    else:
      self.bind('<Control-Key-q>', self.quit)
      self.bind('<Control-Key-w>', self.quit)

  def start(self):
    self.mainloop()

  def quit(self, *args, **kwargs):
    self.destroy()

  def layout(self):
    self.columnconfigure(0, weight=1)
    self.rowconfigure(0, weight=1)

  def create_widgets(self):
    preview_frame = ctk.CTkFrame(self, fg_color='transparent')
    controls_frame = ctk.CTkFrame(self, corner_radius=0)
    self.p = preview_frame
    self.c = controls_frame

    self.debug = tk.BooleanVar(value=DEFAULT_DEBUG)
    self.model = tk.StringVar(value=DEFAULT_MODEL)
    self.prompts_neg = tk.StringVar(value='')

    # The run button is disabled if all the variables aren't defined
    self.run_button = ctk.CTkButton(self.c, text='Run', command=self.on_run, state='disabled')
    self.prompts, self.path = create_required_stringvars(number=2, widget=self.run_button)

    self.model_label = ctk.CTkLabel(self.c, text='Model')
    model_names = [m for m in MODEL_NAME_TO_TYPE]
    self.model_menu = ctk.CTkOptionMenu(self.c, values=model_names, variable=self.model, command=self.on_select_model)
    self.prompts_label = ctk.CTkLabel(self.c, text='Prompts')
    self.prompts_label_neg = ctk.CTkLabel(self.c, text='Negative Prompts')
    self.prompts_entry = ctk.CTkEntry(self.c, textvariable=self.prompts)
    self.prompts_neg = ctk.CTkEntry(self.c)
    self.debug_checkbox = ctk.CTkCheckBox(self.c, text='Debug', variable=self.debug, onvalue=True)

    # Note: CTkLabel doesn't respect text_color_disabled in theme right now
    # ref: https://github.com/TomSchimansky/CustomTkinter/issues/1837
    self.path_label = ctk.CTkLabel(self.p, text='', state='disabled')
    self.preview = ctk.CTkLabel(self.p, text='Drag and drop an input file to preview')
    self.preview.drop_target_register(dnd.DND_ALL)
    self.preview.dnd_bind('<<Drop>>', self.on_drop)

    self.p.grid(row=0, column=0)
    self.p.rowconfigure(0, weight=10)
    self.p.rowconfigure(1, weight=1)
    self.preview.grid(row=0, column=0, sticky=tk.NSEW, padx=0, pady=0)

    self.c.grid(row=1, column=0, sticky=tk.EW)
    self.c.columnconfigure(0, weight=1)
    self.c.columnconfigure(1, weight=10)
    self.prompts_label.grid(row=3, column=0, sticky=tk.E, padx=10, pady=10)
    self.prompts_entry.grid(row=3, column=1, sticky=tk.EW, padx=10, pady=10)
    self.prompts_label_neg.grid(row=4, column=0, sticky=tk.E, padx=10, pady=10)
    self.prompts_neg.grid(row=4, column=1, sticky=tk.EW, padx=10, pady=10)
    self.model_label.grid(row=5, column=0, sticky=tk.E, padx=10, pady=10)
    self.model_menu.grid(row=5, column=1, sticky=tk.EW, padx=10, pady=10)
    self.debug_checkbox.grid(row=6, column=0, sticky=tk.E, padx=10, pady=10)
    self.run_button.grid(row=6, column=1, sticky=tk.EW, padx=10, pady=10)

  def place_path_label(self):
    self.path_label.grid(row=1, column=0, sticky=tk.EW, padx=10, pady=(10, 10))

  def on_select_model(self, model_name):
    log(f'Selected model: {model_name}')

  def on_run(self):
    self.run_button.configure(state='disabled')
    self.config(cursor='watch')
    self.update()
    log('Running job ...')
    model_type = MODEL_NAME_TO_TYPE[self.model.get()]
    model_arg = model_type.replace(HQ_MODEL_PREFIX, '')
    job_args = [
      self.path.get(),
      '--prompts', self.prompts.get(),
      '--model', model_arg,
    ]
    nprompts = self.prompts_neg.get()
    if nprompts and len(nprompts) > 0:
      job_args.extend(['--nprompts', nprompts])
    if HQ_MODEL_PREFIX in model_type:
      job_args.append('--hq')
    if self.debug.get():
      job_args.append('--debug')
    thread = threading.Thread(target=self.job, args=job_args)
    thread.start()

  def job(self, *args):
    debug(f'Calling dummy job {args} ...')
    import time
    time.sleep(2)
    self.on_job_end()

  def on_job_end(self):
    log('Job done!')
    self.config(cursor='')
    # Force update before activating run button again to dispose of any queued click events while button disabled
    self.update()
    self.run_button.configure(state='normal')

  def on_choose_file(self):
    ctk.filedialog.askdirectory()

  def on_drop(self, event):
    path = event.data
    self.path.set(path)
    self.path_label.configure(text=path)
    self.place_path_label()
    try:
      preview_image: pil.Image.Image = get_preview_image(path, App.PREVIEW_WIDTH, App.PREVIEW_HEIGHT)
      ctk_preview = ctk.CTkImage(dark_image=preview_image, size=(App.PREVIEW_WIDTH, App.PREVIEW_HEIGHT))
      self.preview.configure(text='')
      self.preview.configure(image=ctk_preview)
    except Exception as err:
      log(err)


def main(argv=None):
  ctk.set_appearance_mode(APPEARANCE_MODE)
  ctk.set_default_color_theme(resource_path(COLOR_THEME[0]))
  app = App()
  app.start()


if __name__ == '__main__':
  main(sys.argv)



import tkinter as tk
import typing

import customtkinter as ctk

from ezsam.lib.logger import debug


Disableable = typing.TypeVar(
  'Disableable',
  bound=ctk.CTkButton
  | ctk.CTkRadioButton
  | ctk.CTkCheckBox
  | ctk.CTkComboBox
  | ctk.CTkEntry
  | ctk.CTkOptionMenu
  | ctk.CTkSlider
  | ctk.CTkSwitch
  | ctk.CTkTextbox,
)


def create_required_stringvars(number: int, widget: Disableable) -> tk.StringVar:
  """
  Creates a set of variables that on change disable/enable the CTk widget depending if they're all defined.
  """
  tkvars: list[tk.StringVar] = []
  for n in range(number):
    tkvars.append(tk.StringVar(value=''))

  def on_change(name, index, mode):
    all_defined = True
    for value in [var.get() for var in tkvars]:
      debug(f'Value: {value}')
      if not value or len(value) <= 0:
        all_defined = False
        break
    if all_defined:
      debug('All defined, enable widget')
      widget.configure(state='normal')
    else:
      debug('Not all defined, disable widget')
      widget.configure(state='disabled')

  for var in tkvars:
    var.trace_add('write', on_change)
  return tkvars



WIDTH = 480
HEIGHT = 640
PREVIEW_WIDTH = 400
PREVIEW_HEIGHT = PREVIEW_WIDTH
APPEARANCE_MODE = 'system'
COLOR_THEME = [
  'src/ezsam/gui/assets/theme.json',
  'site-packages/ezsam/gui/assets/theme.json',
  # 'blue' # A fallback built-in theme
]

DEFAULT_MODEL = 'SAM Large'
DEFAULT_DEBUG = False



from enum import Enum

from ezsam.cli.config.defaults import DEFAULT_CACHE_FOLDER_LOCATION


class Model(str, Enum):
  vit_h = 'vit_h'
  vit_l = 'vit_l'
  vit_b = 'vit_b'
  hq_vit_h = 'hq_vit_h'
  hq_vit_l = 'hq_vit_l'
  hq_vit_b = 'hq_vit_b'
  hq_vit_tiny = 'hq_vit_tiny'
  gd = 'gd'


MODEL_URL = {
  Model.vit_h: 'https://dl.fbaipublicfiles.com/segment_anything/sam_vit_h_4b8939.pth',
  Model.vit_l: 'https://dl.fbaipublicfiles.com/segment_anything/sam_vit_l_0b3195.pth',
  Model.vit_b: 'https://dl.fbaipublicfiles.com/segment_anything/sam_vit_b_01ec64.pth',
  Model.hq_vit_h: 'https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_h.pth',
  Model.hq_vit_l: 'https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_l.pth',
  Model.hq_vit_b: 'https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_b.pth',
  Model.hq_vit_tiny: 'https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_tiny.pth',
  Model.gd: 'https://github.com/IDEA-Research/GroundingDINO/releases/download/v0.1.0-alpha/groundingdino_swint_ogc.pth',
}

MODEL_FILE_BASENAME = {
  Model.vit_h: 'sam_vit_h_4b8939',
  Model.vit_l: 'sam_vit_l_0b3195',
  Model.vit_b: 'sam_vit_b_01ec64',
  Model.hq_vit_h: 'sam_hq_vit_h',
  Model.hq_vit_l: 'sam_hq_vit_l',
  Model.hq_vit_b: 'sam_hq_vit_b',
  Model.hq_vit_tiny: 'sam_hq_vit_tiny',
  Model.gd: 'groundingdino_swint_ogc',
}

MODEL_SEARCH_LOCATIONS = [
  DEFAULT_CACHE_FOLDER_LOCATION,
  '.',
]


def default_model_locations(model: Model) -> list[str]:
  return [f'{search_path}/models/{MODEL_FILE_BASENAME[model]}.pth' for search_path in MODEL_SEARCH_LOCATIONS]


def get_default_paths_from_model(model: Model) -> str:
  paths = default_model_locations(model)
  if not paths or len(paths) <= 0:
    raise ValueError(f'Invalid model: {model}')
  return paths



import pathlib

from ezsam.cli.formats import OutputImageFormat, OutputVideoCodec

HOME_FOLDER = pathlib.Path.home().as_posix()
DEFAULT_CACHE_FOLDER_LOCATION = f'{HOME_FOLDER}/.cache/ezsam'
DEFAULT_GROUNDING_DINO_CONFIG_PATH = f'{DEFAULT_CACHE_FOLDER_LOCATION}/GroundingDINO_SwinT_OGC.py'
DEFAULT_SAM_MODEL = 'vit_h'
DEFAULT_OUTPUT_DIR = '.'
DEFAULT_OUTPUT_SUFFIX = '.out'
DEFAULT_BOX_THRESHOLD = 0.3
DEFAULT_TEXT_THRESHOLD = 0.3
DEFAULT_NMS_THRESHOLD = 0.8
DEFAULT_IMAGE_FORMAT = OutputImageFormat.png.value
DEFAULT_VIDEO_CODEC = OutputVideoCodec.vp9.value



# Resize an image while keeping aspect ratio
# ref: https://stackoverflow.com/a/44724368
#
# Example: scaled = resize_and_pad(image, (200,200))
#

import cv2
import numpy as np


def resize_and_pad(img, size, pad_color=0):
  h, w = img.shape[:2]
  sh, sw = size

  # interpolation method
  if h > sh or w > sw:  # shrinking image
    interp = cv2.INTER_AREA

  else:  # stretching image
    interp = cv2.INTER_CUBIC

  # aspect ratio of image
  aspect = float(w) / h
  saspect = float(sw) / sh

  if (saspect >= aspect) or ((saspect == 1) and (aspect <= 1)):  # new horizontal image
    new_h = sh
    new_w = np.round(new_h * aspect).astype(int)
    pad_horz = float(sw - new_w) / 2
    pad_left, pad_right = np.floor(pad_horz).astype(int), np.ceil(pad_horz).astype(int)
    pad_top, pad_bot = 0, 0

  elif (saspect < aspect) or ((saspect == 1) and (aspect >= 1)):  # new vertical image
    new_w = sw
    new_h = np.round(float(new_w) / aspect).astype(int)
    pad_vert = float(sh - new_h) / 2
    pad_top, pad_bot = np.floor(pad_vert).astype(int), np.ceil(pad_vert).astype(int)
    pad_left, pad_right = 0, 0

  # set pad color
  if len(img.shape) == 3 and not isinstance(
    pad_color, (list, tuple, np.ndarray)
  ):  # color image but only one color provided
    pad_color = [pad_color] * 3

  # scale and pad
  scaled_img = cv2.resize(img, (new_w, new_h), interpolation=interp)
  scaled_img = cv2.copyMakeBorder(
    scaled_img, pad_top, pad_bot, pad_left, pad_right, borderType=cv2.BORDER_CONSTANT, value=pad_color
  )

  return scaled_img



# Abstract paths to work in both development and inside of binaries.
# Needed only for non-code files, for example JSON.

import os
import sys
import tempfile

from ezsam.lib.config import EXECUTABLE_NAME
from ezsam.lib.logger import debug, warn


# ref: https://stackoverflow.com/a/13790741
def resource_path(relative_path: str, throw_error = False) -> str:
  """
  Try to get absolute path to resource, whether in development or in standalone/one-file executables for Nuitka or PyInstaller.
  """
  # Note:
  # - PyInstaller creates a temp folder and stores that path in sys._MEIPASS
  # - Nuitka build creates temp folder at --onefile-tempdir-spec=$tempdir
  #    The default tempdir is: {TEMP}/onefile_{PID}_{TIME}
  #    For this project's build it's set to: {TEMP}/ezsam
  #    Another example, using persistence: {CACHE_DIR}/{PRODUCT}/{VERSION}
  try:
    temp = tempfile.gettempdir()
    base_path = f'{temp}/{EXECUTABLE_NAME}'
    path = os.path.normpath(os.path.join(base_path, relative_path))
    open(path)
  except FileNotFoundError:
    base_path = getattr(sys, '_MEIPASS', os.path.dirname(__file__) + '/../../../')
    try:
      path = os.path.normpath(os.path.join(base_path, relative_path))
      open(path)
    except FileNotFoundError as err:
      if throw_error:
        raise err
      warn(f'Could not find file at {relative_path}, returning {path} anyways ...')
  debug(f'Converted {relative_path} to {path}')
  return path
